@@ -5,6 +5,7 @@ use holo_hash::HeaderHash;
 use holochain_sqlite::rusqlite::Transaction;
 use holochain_types::dht_op::produce_op_lights_from_elements;
 use holochain_types::dht_op::produce_op_lights_from_iter;
+use holochain_types::dht_op::DhtOp;
 use holochain_types::dht_op::DhtOpLight;
 use holochain_types::dht_op::DhtOpType;
 use holochain_types::dht_op::OpOrder;
@@ -13,11 +14,14 @@ use holochain_types::element::SignedHeaderHashedExt;
 use holochain_types::env::EnvRead;
 use holochain_types::env::EnvWrite;
 use holochain_types::timestamp;
+use holochain_types::Timestamp;
 use holochain_zome_types::entry::EntryHashed;
 use holochain_zome_types::header;
 use holochain_zome_types::CapAccess;
 use holochain_zome_types::CapGrant;
 use holochain_zome_types::CapSecret;
+use holochain_zome_types::CounterSigningAgentState;
+use holochain_zome_types::CounterSigningSessionData;
 use holochain_zome_types::Element;
 use holochain_zome_types::Entry;
 use holochain_zome_types::EntryVisibility;
@@ -27,15 +31,19 @@ use holochain_zome_types::HeaderBuilder;
 use holochain_zome_types::HeaderBuilderCommon;
 use holochain_zome_types::HeaderHashed;
 use holochain_zome_types::HeaderInner;
+use holochain_zome_types::PreflightRequest;
 use holochain_zome_types::QueryFilter;
 use holochain_zome_types::Signature;
 use holochain_zome_types::SignedHeader;
 use holochain_zome_types::SignedHeaderHashed;
 
+use crate::chain_lock::is_chain_locked;
+use crate::chain_lock::is_lock_expired;
 use crate::prelude::*;
 use crate::query::chain_head::ChainHeadQuery;
 use crate::scratch::Scratch;
 use crate::scratch::SyncScratch;
+use holo_hash::EntryHash;
 use holochain_serialized_bytes::prelude::*;
 
 pub use error::*;
@@ -48,6 +56,7 @@ pub struct SourceChain {
     author: Arc<AgentPubKey>,
     persisted_seq: u32,
     persisted_head: HeaderHash,
+    persisted_timestamp: Timestamp,
     public_only: bool,
 }
 
@@ -73,7 +82,7 @@ impl SourceChain {
     pub async fn new(vault: EnvWrite, author: AgentPubKey) -> SourceChainResult<Self> {
         let scratch = Scratch::new().into_sync();
         let author = Arc::new(author);
-        let (persisted_head, persisted_seq) = vault
+        let (persisted_head, persisted_seq, persisted_timestamp) = vault
             .async_reader({
                 let author = author.clone();
                 move |txn| chain_head_db(&txn, author)
@@ -85,6 +94,7 @@ impl SourceChain {
             author,
             persisted_seq,
             persisted_head,
+            persisted_timestamp,
             public_only: false,
         })
     }
@@ -110,47 +120,73 @@ impl SourceChain {
     // TODO: Maybe we should store data as elements in the scratch?
     // TODO: document that this is only the elemnts in the SCRATCH, not the
     //       entire source chain!
-    pub fn elements(&self) -> SourceChainResult<Vec<Element>> {
+    pub fn scratch_elements(&self) -> SourceChainResult<Vec<Element>> {
         Ok(self.scratch.apply(|scratch| scratch.elements().collect())?)
     }
 
-    pub fn chain_head(&self) -> SourceChainResult<(HeaderHash, u32)> {
+    pub fn chain_head(&self) -> SourceChainResult<(HeaderHash, u32, Timestamp)> {
         // Check scratch for newer head.
         Ok(self.scratch.apply(|scratch| {
             let chain_head = chain_head_scratch(&(*scratch), self.author.as_ref());
-            let (prev_header, header_seq) =
-                chain_head.unwrap_or_else(|| (self.persisted_head.clone(), self.persisted_seq));
-            (prev_header, header_seq)
+            let (prev_header, header_seq, timestamp) = chain_head.unwrap_or_else(|| {
+                (
+                    self.persisted_head.clone(),
+                    self.persisted_seq,
+                    self.persisted_timestamp,
+                )
+            });
+            (prev_header, header_seq, timestamp)
         })?)
     }
 
+    async fn put_with_header(
+        &self,
+        header: Header,
+        maybe_entry: Option<Entry>,
+    ) -> SourceChainResult<HeaderHash> {
+        let header = HeaderHashed::from_content_sync(header);
+        let hash = header.as_hash().clone();
+        let header = SignedHeaderHashed::new(&self.vault.keystore(), header).await?;
+        let element = Element::new(header, maybe_entry);
+        self.scratch
+            .apply(|scratch| insert_element_scratch(scratch, element))?;
+        Ok(hash)
+    }
+
+    pub async fn put_countersigned(&self, entry: Entry) -> SourceChainResult<HeaderHash> {
+        let entry_hash = EntryHash::with_data_sync(&entry);
+        if let Entry::CounterSign(ref session_data, _) = entry {
+            self.put_with_header(
+                Header::from_countersigning_data(entry_hash, session_data, (*self.author).clone())?,
+                Some(entry),
+            )
+            .await
+        } else {
+            // The caller MUST guard against this case.
+            unreachable!("Put countersigned called with the wrong entry type");
+        }
+    }
+
     pub async fn put<H: HeaderInner, B: HeaderBuilder<H>>(
         &self,
         header_builder: B,
         maybe_entry: Option<Entry>,
     ) -> SourceChainResult<HeaderHash> {
-        let (prev_header, chain_head_seq) = self.chain_head()?;
+        let (prev_header, chain_head_seq, chain_head_timestamp) = self.chain_head()?;
         let header_seq = chain_head_seq + 1;
 
         // Build the header.
         let common = HeaderBuilderCommon {
             author: (*self.author).clone(),
-            timestamp: timestamp::now(),
+            timestamp: std::cmp::max(
+                timestamp::now(),
+                (chain_head_timestamp + std::time::Duration::from_nanos(1))?,
+            ),
             header_seq,
             prev_header,
         };
-        let header = header_builder.build(common).into();
-        let header = HeaderHashed::from_content_sync(header);
-        let hash = header.as_hash().clone();
-
-        // Sign the header.
-        let header = SignedHeaderHashed::new(&self.vault.keystore(), header).await?;
-        let element = Element::new(header, maybe_entry);
-
-        // Put into scratch.
-        self.scratch
-            .apply(|scratch| insert_element_scratch(scratch, element))?;
-        Ok(hash)
+        self.put_with_header(header_builder.build(common).into(), maybe_entry)
+            .await
     }
 
     pub fn has_initialized(&self) -> SourceChainResult<bool> {
@@ -164,7 +200,8 @@ impl SourceChain {
     #[allow(clippy::len_without_is_empty)]
     pub fn len(&self) -> SourceChainResult<u32> {
         Ok(self.scratch.apply(|scratch| {
-            let scratch_max = chain_head_scratch(&(*scratch), self.author.as_ref()).map(|(_, s)| s);
+            let scratch_max =
+                chain_head_scratch(&(*scratch), self.author.as_ref()).map(|(_, s, _)| s);
             scratch_max
                 .map(|s| std::cmp::max(s, self.persisted_seq))
                 .unwrap_or(self.persisted_seq)
@@ -381,6 +418,93 @@ impl SourceChain {
         Ok(elements)
     }
 
+    pub async fn unlock_chain(&self) -> SourceChainResult<()> {
+        self.vault
+            .async_commit(move |txn| unlock_chain(txn))
+            .await?;
+        Ok(())
+    }
+
+    pub async fn is_chain_locked(&self, lock: Vec<u8>) -> SourceChainResult<bool> {
+        Ok(self
+            .vault
+            .async_reader(move |txn| is_chain_locked(&txn, &lock))
+            .await?)
+    }
+
+    pub async fn accept_countersigning_preflight_request(
+        &self,
+        preflight_request: PreflightRequest,
+        agent_index: u8,
+    ) -> SourceChainResult<CounterSigningAgentState> {
+        let hashed_preflight_request = holo_hash::encode::blake2b_256(
+            &holochain_serialized_bytes::encode(&preflight_request)?,
+        );
+
+        // This all needs to be ensured in a non-panicky way BEFORE calling into the source chain here.
+        let author = self.author.clone();
+        assert_eq!(
+            *author,
+            preflight_request.signing_agents()[agent_index as usize].0
+        );
+
+        let countersigning_agent_state = self
+            .vault
+            .async_commit(move |txn| {
+                if is_chain_locked(txn, &hashed_preflight_request)? {
+                    return Err(SourceChainError::ChainLocked);
+                }
+                let (persisted_head, persisted_seq, _) = chain_head_db(&txn, author)?;
+                let countersigning_agent_state =
+                    CounterSigningAgentState::new(agent_index, persisted_head, persisted_seq);
+                lock_chain(
+                    txn,
+                    &hashed_preflight_request,
+                    preflight_request.session_times().end(),
+                )?;
+                SourceChainResult::Ok(countersigning_agent_state)
+            })
+            .await?;
+        Ok(countersigning_agent_state)
+    }
+
+    /// If there is a countersigning session get the
+    /// StoreEntry op to send to the entry authorities.
+    pub fn countersigning_op(&self) -> SourceChainResult<Option<DhtOp>> {
+        let r = self.scratch.apply(|scratch| {
+            scratch
+                .entries()
+                .find(|e| matches!(**e.1, Entry::CounterSign(_, _)))
+                .and_then(|(entry_hash, entry)| {
+                    scratch
+                        .headers()
+                        .find(|shh| {
+                            shh.header()
+                                .entry_hash()
+                                .map(|eh| eh == entry_hash)
+                                .unwrap_or(false)
+                        })
+                        .and_then(|shh| {
+                            Some(DhtOp::StoreEntry(
+                                shh.signature().clone(),
+                                shh.header().clone().try_into().ok()?,
+                                Box::new((**entry).clone()),
+                            ))
+                        })
+                })
+        })?;
+        Ok(r)
+    }
+
+    pub fn lock_for_entry(entry: Option<&Entry>) -> SourceChainResult<Vec<u8>> {
+        Ok(match entry {
+            Some(Entry::CounterSign(session_data, _)) => holo_hash::encode::blake2b_256(
+                &holochain_serialized_bytes::encode(session_data.preflight_request())?,
+            ),
+            _ => Vec::with_capacity(0),
+        })
+    }
+
     pub async fn flush(&self) -> SourceChainResult<()> {
         // Nothing to write
         if self.scratch.apply(|s| s.is_empty())? {
@@ -439,13 +563,24 @@ impl SourceChain {
             SourceChainResult::Ok((headers, ops, entries))
         })?;
 
+        let maybe_countersigned_entry = entries
+            .iter()
+            .map(|entry| entry.as_content())
+            .find(|entry| matches!(entry, Entry::CounterSign(_, _)));
+
+        if matches!(maybe_countersigned_entry, Some(Entry::CounterSign(_, _))) && headers.len() != 1
+        {
+            return Err(SourceChainError::DirtyCounterSigningWrite);
+        }
+        let lock = Self::lock_for_entry(maybe_countersigned_entry)?;
+
         // Write the entries, headers and ops to the database in one transaction.
         let author = self.author.clone();
         let persisted_head = self.persisted_head.clone();
         self.vault
-            .async_commit(move |txn| {
+            .async_commit(move |txn: &mut Transaction| {
                 // As at check.
-                let (new_persisted_head, _) = chain_head_db(&txn, author)?;
+                let (new_persisted_head, _, _) = chain_head_db(&txn, author)?;
                 if headers.last().is_none() {
                     // Nothing to write
                     return Ok(());
@@ -457,6 +592,22 @@ impl SourceChain {
                     ));
                 }
 
+                // If the lock isn't empty this is a countersigning session.
+                let is_countersigning_session = !lock.is_empty();
+
+                if is_chain_locked(txn, &lock)? {
+                    return Err(SourceChainError::ChainLocked);
+                }
+                // If this is a countersigning session, and the chain is NOT
+                // locked then either the session expired or the countersigning
+                // entry being committed now is the correct one for the lock.
+                else if is_countersigning_session {
+                    // If the lock is expired then we can't write this countersigning session.
+                    if is_lock_expired(txn, &lock)? {
+                        return Err(SourceChainError::LockExpired);
+                    }
+                }
+
                 for entry in entries {
                     insert_entry(txn, entry)?;
                 }
@@ -476,15 +627,22 @@ impl SourceChain {
                     // StoreEntry ops with private entries are never gossiped or published
                     // so we don't need to integrate them.
                     // TODO: Can anything every depend on a private store entry op? I don't think so.
-                    if !(op_type == DhtOpType::StoreEntry
-                        && visibility == Some(EntryVisibility::Private))
-                    {
+                    let is_private_entry = op_type == DhtOpType::StoreEntry
+                        && visibility == Some(EntryVisibility::Private);
+
+                    // Don't publish private entries or countersigning sessions.
+                    if !is_private_entry && !is_countersigning_session {
                         set_validation_stage(
                             txn,
-                            op_hash,
+                            op_hash.clone(),
                             ValidationLimboStatus::AwaitingIntegration,
                         )?;
                     }
+                    // If this is a countersigning session we want to withhold
+                    // publishing the ops until the session is successful.
+                    if is_countersigning_session {
+                        set_withhold_publish(txn, op_hash)?;
+                    }
                 }
                 SourceChainResult::Ok(())
             })
@@ -604,20 +762,27 @@ pub fn put_raw(
 fn chain_head_db(
     txn: &Transaction,
     author: Arc<AgentPubKey>,
-) -> SourceChainResult<(HeaderHash, u32)> {
+) -> SourceChainResult<(HeaderHash, u32, Timestamp)> {
     let chain_head = ChainHeadQuery::new(author);
-    let (prev_header, last_header_seq) = chain_head
+    let (prev_header, last_header_seq, last_header_timestamp) = chain_head
         .run(Txn::from(txn))?
         .ok_or(SourceChainError::ChainEmpty)?;
-    Ok((prev_header, last_header_seq))
+    Ok((prev_header, last_header_seq, last_header_timestamp))
 }
 
-fn chain_head_scratch(scratch: &Scratch, author: &AgentPubKey) -> Option<(HeaderHash, u32)> {
+fn chain_head_scratch(
+    scratch: &Scratch,
+    author: &AgentPubKey,
+) -> Option<(HeaderHash, u32, Timestamp)> {
     scratch
         .headers()
         .filter_map(|shh| {
             if shh.header().author() == author {
-                Some((shh.header_address().clone(), shh.header().header_seq()))
+                Some((
+                    shh.header_address().clone(),
+                    shh.header().header_seq(),
+                    shh.header().timestamp(),
+                ))
             } else {
                 None
             }
@@ -625,6 +790,39 @@ fn chain_head_scratch(scratch: &Scratch, author: &AgentPubKey) -> Option<(Header
         .max_by_key(|h| h.1)
 }
 
+/// Check if there is a current countersigning session and if so, return the
+/// session data and the entry hash.
+pub fn current_countersigning_session(
+    txn: &Transaction<'_>,
+    author: Arc<AgentPubKey>,
+) -> SourceChainResult<Option<(EntryHash, CounterSigningSessionData)>> {
+    // The chain must be locked for a session to be active.
+    if is_chain_locked(txn, &[])? {
+        match chain_head_db(txn, author) {
+            // We haven't done genesis so no session can be active.
+            Err(SourceChainError::ChainEmpty) => Ok(None),
+            Err(e) => Err(e),
+            Ok((hash, _, _)) => {
+                let txn: Txn = txn.into();
+                // Get the session data from the database.
+                let element = match txn.get_element(&hash.into())? {
+                    Some(element) => element,
+                    None => return Ok(None),
+                };
+                let (shh, ee) = element.into_inner();
+                Ok(match (shh.header().entry_hash(), ee.into_option()) {
+                    (Some(entry_hash), Some(Entry::CounterSign(cs, _))) => {
+                        Some((entry_hash.to_owned(), *cs))
+                    }
+                    _ => None,
+                })
+            }
+        }
+    } else {
+        Ok(None)
+    }
+}
+
 #[cfg(test)]
 async fn _put_db<H: HeaderInner, B: HeaderBuilder<H>>(
     vault: holochain_types::env::EnvWrite,
@@ -632,7 +830,7 @@ async fn _put_db<H: HeaderInner, B: HeaderBuilder<H>>(
     header_builder: B,
     maybe_entry: Option<Entry>,
 ) -> SourceChainResult<HeaderHash> {
-    let (prev_header, last_header_seq) =
+    let (prev_header, last_header_seq, _) =
         fresh_reader!(vault, |txn| { chain_head_db(&txn, author.clone()) })?;
     let header_seq = last_header_seq + 1;
 
@@ -651,7 +849,7 @@ async fn _put_db<H: HeaderInner, B: HeaderBuilder<H>>(
     let entry = entry.into_option();
     let hash = header.as_hash().clone();
     vault.conn()?.with_commit_sync(|txn| {
-        let (new_head, _) = chain_head_db(txn, author.clone())?;
+        let (new_head, _, _) = chain_head_db(txn, author.clone())?;
         if new_head != prev_header {
             return Err(SourceChainError::HeadMoved(
                 Some(prev_header),
